# Pairs of interlisted USD/CAD stocks and ETFs for Norbert's Gambit
#
# Mostly taken from "Best stocks for Norbert's Gambit" thread:
# https://www.canadianmoneyforum.com/threads/dual-listed-etfs-tsx-nyse.135364/post-1972456

ng_pairs = (                              # US$ symbol    CA$ symbol
    ('Horizons U.S. Dollar Currency ETF', 'DLR-U.TO',    'DLR.TO'),
    ('TD (Canadian bank)',                'TD',          'TD.TO'),
    ('BMO (Canadian bank)',               'BMO',         'BMO.TO'),
    ('CIBC (Canadian bank)',              'CM',          'CM.TO'),
    ('ScotiaBank (Canadian bank)',        'BNS',         'BNS.TO'),
    ('RBC (Canadian bank)',               'RY',          'RY.TO'),
    ('Canadian National Railway',         'CNI',         'CNR.TO'),
    ('Enbridge (oil/energy)',             'ENB',         'ENB.TO'),
    ('Suncor (oil/energy)',               'SU',          'SU.TO'),
    ('MFC (insurance/investment)',        'MFC',         'MFC.TO'),
    ('Horizons S&P 500 ETF',              'HXS-U.TO',    'HXS.TO'),
    ('Horizons TSX60 ETF',                'HXT-U.TO',    'HXT.TO'),
    ('Horizons Global Dev Index ETF',     'HXDM-U.TO',   'HXDM.TO'),
)

# IT IS NOT POSSIBLE TO USE THESE PAIRS FOR NORBERT'S GAMBIT
# because the USD/CAD symbols do not share the same CUSIPs.
#
# Taken from https://www.finiki.org/wiki/Norbert%27s_gambit#ETFs_with_different_CUSIPs

bad_list = (
    'ZSP.U', 'ZSP',
    'XEF.U', 'XEF',
    'XUS.U', 'XUS',
    'XUU.U', 'XUU',
)



import argparse
import logging
import urllib.parse
import json
from datetime import datetime, timezone, timedelta
import time
from sys import stdout
import csv
import os
from itertools import count
#from dataclasses import dataclass

import math
from inspect import isfunction, isbuiltin
all_math_funcs = {n: f for (n, f) in vars(math).items() if isfunction(f) or isbuiltin(f)}

import requests
from colored import Fore, Back, Style

from .pairs import ng_pairs, bad_list
from .yq import YFQuote

logging.basicConfig(level=os.environ.get('LOGLEVEL', 'WARNING').strip().upper())
logging.getLogger('urllib3').setLevel(logging.WARNING)
logger = logging.getLogger(__name__)


# Quacks like a dict and an object
class QuotePair(dict):
    @property
    def src(self):
        return self.USD if self.src_cur == 'USD' else self.CAD
    @property
    def dst(self):
        return self.CAD if self.src_cur == 'USD' else self.USD

    def __getattr__(self, k):
        try:
            return self[k]
        except KeyError as e:
            raise AttributeError(*e.args)
    def __setattr__(self, k, v):
        self[k]=v


def get_ng_data(src_cur: str, yfq: YFQuote):
    global ng_pairs
    j = {}
    for desc, usd, cad in ng_pairs:
        assert usd not in bad_list
        assert cad not in bad_list
        for currency, symbol in (('USD', usd), ('CAD', cad)):
            j[desc] = QuotePair(src_cur=src_cur,
                                USD=yfq.get_quote(desc, usd, 'USD'),
                                CAD=yfq.get_quote(desc, cad, 'CAD'))
    return j


def main():
    p = argparse.ArgumentParser(description="A tool to help you pick the optimal securities to convert between US and Canadian currency using Norbert's Gambit, based on near-realtime quotes from Yahoo Finance.")
    p.add_argument('-v', '--verbose', action='count', default=0, help='Show long results with full calculations')
    p.add_argument('--max-lag', type=int, default=60, help='Maximum lag to accept (in seconds)')
    p.add_argument('-L', '--limit', type=int, default=math.inf, help='Only show the best LIMIT results')
    p.add_argument('src_cur', type=str.upper, choices=('USD', 'CAD'), help='Source currency (USD or CAD)')
    p.add_argument('src_amount', type=float, help='Amount of source currency to convert')
    g = p.add_argument_group('Commissions', '''
        These may be either fixed strings, or Python expressions using the variables `src_ask`, `dst_bid`, `shares`, `src_amount_convert`, and `dst_amount`,
        as well as any functions from `math` (e.g. `floor`) or builtins (e.g. `max`).''')
    g.add_argument('-S', '--src-commission', type=str, default='6.95', help="Commission for buying source-currency security (default %(default)r)")
    g.add_argument('-D', '--dst-commission', type=str, default='6.95', help="Commission for selling destination-currency security (default %(default)r)")

    args = p.parse_args()

    src_amount = args.src_amount
    src_cur = args.src_cur
    dst_cur = 'CAD' if src_cur == 'USD' else 'USD'

    now = time.time()

    yfq = YFQuote()
    j = get_ng_data(src_cur, yfq)
    mmex = yfq.get_quote('USD/CAD mid-market rate', 'CAD=X', 'CAD')
    mm_rate = mmex.last_price ** (+1 if src_cur == 'USD' else -1)
    mm_lag = now - mmex.timestamp
    if mm_lag > args.max_lag:
        p.error(f'Lag in London mid-market exchange rate is too high ({mm_lag:.0f} sec). Are US and Canadian markets open?')

    print(f"Finding optimal securities to convert {Fore.red}{src_cur} {src_amount:,.02f}{Style.reset} to {Fore.green}{dst_cur}{Style.reset} using Norbert's Gambit.")
    print(f'- Commission function for buying {src_cur} security:  {Fore.red}{args.src_commission}{Style.reset}')
    print(f'- Commission function for selling {dst_cur} security: {Fore.green}{args.dst_commission}{Style.reset}')
    print(f'\nLondon mid-market exchange rate for {Fore.red}{src_cur}{Style.reset} -> {Fore.green}{dst_cur}{Style.reset}'
          f' is {Fore.yellow}{mm_rate:,.04f}{Style.reset}'
          f' ({Style.bold}{mm_lag:.0f} sec lag{Style.reset})\n')

    # Calculate results and stash them in the dict of quotes
    for desc, jd in j.items():
        for reduce in count():
            # How many shares should we buy/sell?
            shares = jd['shares'] = src_amount // jd.src.ask - reduce
            src_amount_convert = src_amount_convert = shares * jd.src.ask
            dst_amount = shares * jd.dst.bid

            # Calculate commissions
            commission_vars = dict(all_math_funcs, src_ask=jd.src.ask, dst_bid=jd.dst.bid, shares=shares, src_amount_convert=src_amount_convert, dst_amount=dst_amount)
            src_commission = jd['src_commission'] = eval(args.src_commission, commission_vars)
            dst_commission = jd['dst_commission'] = eval(args.dst_commission, commission_vars)

            # Net amounts outgoing and incoming after commissions
            src_amount_net = jd['src_amount_net'] = src_amount_convert + src_commission
            src_leftover = jd['src_leftover'] = src_amount - src_amount_net
            dst_amount_net = jd['dst_amount_net'] = dst_amount - dst_commission

            # Keep reducing by one share until we're below the starting amount after commission
            if src_leftover >= 0.0:
                break

        # Stash results
        jd['effective_rate'] = dst_amount_net / src_amount_net
        jd['theoretical_rate'] = jd.dst.bid / jd.src.ask
        jd['src_lag'] = now - jd.src.timestamp
        jd['dst_lag'] = now - jd.dst.timestamp

    lag_ok = [(desc, jd) for (desc, jd) in j.items() if jd['src_lag'] <= args.max_lag and jd['dst_lag'] <= args.max_lag]
    if not lag_ok:
        p.error(f'Out of {len(j)} interlisted stocks/ETFs, none have lag of <={args.max_lag} sec. Are US and Canadian markets open?')
    lag_ok.sort(key=lambda x: x[1]['effective_rate'], reverse=True)
    print(f'Out of {len(j)} pairs of interlisted stocks/ETFs, {len(lag_ok)} have lag of <={args.max_lag} sec.')

    # Display results ranked from best to worst
    print('Best options:\n')
    for ii, (desc, jd) in enumerate(lag_ok):
        if ii >= args.limit:
            break

        shares = jd['shares']
        src_commission = jd['src_commission']
        dst_commission = jd['dst_commission']
        src_amount_net = jd['src_amount_net']
        src_leftover = jd['src_leftover']
        dst_amount_net = jd['dst_amount_net']
        effective_rate = jd['effective_rate']
        theoretical_rate = jd['theoretical_rate']
        src_lag = jd['src_lag']
        dst_lag = jd['dst_lag']

        dst_amount_mm = src_amount_net * mm_rate
        loss_compared_to_mm = dst_amount_mm - dst_amount_net

        if args.verbose < 2:
            print(f'{ii+1:-2d}. Buy {Style.bold}{shares:.0f}{Style.reset} x {Fore.red}{jd.src.symbol}{Style.reset} at {Fore.red}{src_cur} {jd.src.ask:,.03f}{Style.reset} ({Style.bold}{src_lag:.0f} sec lag{Style.reset}),'
                  f' sell {Fore.green}{jd.dst.symbol}{Style.reset} at {Fore.green}{dst_cur} {jd.dst.bid:,.03f}{Style.reset} ({Style.bold}{dst_lag:.0f} sec lag{Style.reset})\n'
                  f'    Effective rate of {Fore.yellow}{effective_rate:.04f}{Style.reset}\n'
                  f'    Losing {Fore.green}{dst_cur} {loss_compared_to_mm:,.04f}{Style.reset} compared to London mid-market')
            if args.verbose > 0:
                  print(f'    Net of commissions of {Fore.red}{src_cur} {src_commission:,.02f}{Style.reset} (buy) and {Fore.green}{dst_cur} {dst_commission:,.02f}{Style.reset} (sell)')
            print()

        else:
            print(f'{ii+1:-2d}. Using {desc}:       [ {Style.bold}{max(src_lag, dst_lag):.0f} sec data lag{Style.reset} ]\n\n'
                f'    a. Buy {Style.bold}{shares:.0f}{Style.reset} shares of {Fore.red}{jd.src.symbol}{Style.reset} at ask of {Fore.red}{src_cur} {jd.src.ask:,.03f}{Style.reset}, plus {Fore.red}{src_cur} {src_commission}{Style.reset} commission\n'
                f'       (= {shares:.0f} x {jd.src.ask:,.03f} + {src_commission:,.02f} = {src_amount_net:,.02f})\n'
                f'    b. Sell {Style.bold}{shares:.0f}{Style.reset} shares of {Fore.green}{jd.dst.symbol}{Style.reset} at bid of {Fore.green}{dst_cur} {jd.dst.bid:,.03f}{Style.reset}, less {Fore.green}{dst_cur} {dst_commission}{Style.reset} commission\n'
                f'       (= {shares:.0f} x {jd.dst.bid:,.03f} - {dst_commission:,.02f} = {dst_amount_net:,.02f})\n\n'
                f'    You end up with: {Fore.green}{dst_cur} {dst_amount_net:,.02f}{Style.reset} (+ leftover {Fore.red}{src_cur} {src_leftover:,.02f}{Style.reset})\n'
                f'    Your effective conversion rate: {Fore.yellow}{effective_rate:.04f}{Style.reset}\n'
                f'    Mid-market conversion rate:     {Fore.yellow}{mm_rate:.04f}{Style.reset}\n'
                f'    Compared to MM rate, you lose:  {Fore.green}{dst_cur} {loss_compared_to_mm:,.04f}{Style.reset}\n')

if __name__ == '__main__':
    main()

    #print(dst_bid_over_src_ask)
    #print(eff_rate)
    #print(f'{src_cur} {src_amount} -> {dst_cur} {dst_amount_net}')

    #for currency, jc in jd.items():
        #wr.writerow((jc['symbol'], desc, jc['bid_size'], jc['bid'], jc['ask'], jc['ask_size'], jc['last_price'], jc['change'], jc['change_percent'], now - jc['timestamp']))

# with open('/tmp/foo.json', 'w') as outf:
#    json.dump(j, outf)



from datetime import timezone, timedelta, time
import logging
from urllib.parse import quote_plus
from typing import Optional
from dataclasses import dataclass

import requests

logger = logging.getLogger(__name__)


@dataclass
class YFQuoteResult:
    symbol: str
    currency: str
    timestamp: int
    tz: timezone
    bid_size: Optional[int]
    ask_size: Optional[int]
    bid: Optional[float]
    ask: Optional[float]
    low: Optional[float]
    high: Optional[float]
    last_price: Optional[float]
    change: Optional[float]
    change_percent: Optional[float]


def nav(r, *path, types_ok=(float, int), converter=None, ignore=(), expl='r'):
    for key in path:
        if not isinstance(r, dict):
            logger.warning(f'{expl} is unexpectedly of type {type(r)} rather than dict, returning None')
            return None
        elif key not in r:
            logger.warning(f'{expl} unexpectedly lacks key {key!r}, returning None')
            return None
        r = r[key]
        expl += f'[{key!r}]'
    if not(isinstance(r, types_ok)):
        logger.warning(f'{expl} is unexpectedly of type {type(r)} rather than {" OR ".join(types_ok)}, returning None')
        return None
    elif r in ignore:
        logger.warning(f'{expl} has value {r} which we ignore, returning None')
        return None

    return converter(r) if converter else r


class YFQuote:
    def __init__(self, sess: Optional[requests.Session]=None, crumb: Optional[str]=None, cookies: Optional[dict]=None):
        if sess is None:
            sess = requests.session()
            sess.headers.update({
                'User-Agent': 'Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:129.0) Gecko/20100101 Firefox/129.0',
                'Accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8',
                #'Accept-Encoding': 'gzip, deflate',
                #'Accept-Language': 'en-US,en;q=0.5',
                #'Connection': 'keep-alive',
                #'Cache-Control': 'max-age=0',
            })

        if crumb is None or cookies is None:
            r = sess.get('https://query1.finance.yahoo.com/v1/test/getcrumb')
            r.raise_for_status()
            crumb = r.text
            logger.info(f"Got Yahoo Finance crumb {crumb!r} and cookies {','.join(c.name for c in sess.cookies if ('.'+c.domain).endswith('.yahoo.com'))}")

        self.sess = sess
        self.crumb = crumb

    def get_quote(self, desc, symbol, currency):
        r = self.sess.get(f'https://query1.finance.yahoo.com/v10/finance/quoteSummary/{symbol}?formatted=false&'
                    'modules=quoteType,summaryDetail,price'
                    # More available: ',summaryProfile,financialData,recommendationTrend,earnings,equityPerformance,defaultKeyStatistics,calendarEvents,esgScores,pageViews,financialsTemplate'
                    f'&lang=en-US&region=US&crumb={quote_plus(self.crumb)}')
        r.raise_for_status()

        jsym = r.json()
        jdesc = f'JSON for symbol {symbol} ({currency} side of {desc} pair)'
        if (err := jsym.get('error')) is not None:
            raise RuntimeError(f'Got error {err!r} for {jdesc}')
        if (qs := jsym.get('quoteSummary')) is None:
            raise AssertionError(f'No quoteSummary in {jdesc}')
        if len(res := qs.get('result', ())) != 1:
            raise AssertionError(f'quoteSummary.result has length {len(res)} rather than expected 1 in {jdesc}')

        res, = res
        expl = jdesc + ': quoteSummary.result[0]'

        assert (c := nav(res, 'summaryDetail', 'currency', types_ok=str)) == currency, \
            f'quoteSummary.result[0].summaryDetail.currency is {c!r} rather than expected {currency!r} in {jdesc}'
        # FIXME: USDCAD=X has a symbol of CAD=X here in the JSON...?
        assert (s := nav(res, 'quoteType', 'symbol', types_ok=str)) == symbol or (symbol == 'USDCAD=X' and s == 'CAD=X'), \
        f'quoteSummary.result[0].quoteType.symbol is {s!r} rather than expected {symbol!r} in {jdesc}'

        tzoffset = nav(res, 'quoteType', 'gmtOffSetMilliseconds')
        tzname = nav(res, 'quoteType', 'timeZoneFullName', types_ok=str)
        assert tzoffset is not None and tzname is not None
        tz = timezone(timedelta(seconds=tzoffset / 1000), tzname)

        q = YFQuoteResult(
            symbol = symbol, tz = tz, currency = currency,
            timestamp = nav(res, 'price', 'regularMarketTime', expl=expl), # isoformat, converter=lambda ts: datetime.fromtimestamp(ts, tz), expl=expl).isoformat(),
            # FIXME: why are bid/ask size always zero for TSX symbols? https://www.reddit.com/r/YAHOOFINANCE/comments/1fahk77
            bid_size = nav(res, 'summaryDetail', 'bidSize', expl=expl, ignore=(0,)),
            ask_size = nav(res, 'summaryDetail', 'askSize', expl=expl, ignore=(0,)),
            bid = nav(res, 'summaryDetail', 'bid', expl=expl),
            ask = nav(res, 'summaryDetail', 'ask', expl=expl),
            low = nav(res, 'price', 'regularMarketDayLow', expl=expl),
            high = nav(res, 'price', 'regularMarketDayHigh', expl=expl),
            last_price = nav(res, 'price', 'regularMarketPrice', expl=expl),
            change = nav(res, 'price', 'regularMarketChange', expl=expl),
            change_percent = nav(res, 'price', 'regularMarketChangePercent', expl=expl),
        )
        logger.info(f'Got {jdesc}.')
        return q
